@@ -33,19 +33,26 @@
 #include "ofono.h"
 
 #include "common.h"
+#include "storage.h"
 
 #define DATA_CONNECTION_MANAGER_INTERFACE "org.ofono.DataConnectionManager"
 #define DATA_CONTEXT_INTERFACE "org.ofono.PrimaryDataContext"
 
 #define GPRS_FLAG_ATTACHING 0x1
 
+#define SETTINGS_STORE "gprs"
+#define SETTINGS_GROUP "Settings"
+#define MAX_CONTEXT_NAME_LENGTH 127
+#define MAX_CONTEXTS 256
+
 static GSList *g_drivers = NULL;
 static GSList *g_context_drivers = NULL;
 
 enum gprs_context_type {
-	GPRS_CONTEXT_TYPE_INTERNET = 1,
+	GPRS_CONTEXT_TYPE_INTERNET = 0,
 	GPRS_CONTEXT_TYPE_MMS,
 	GPRS_CONTEXT_TYPE_WAP,
+	GPRS_CONTEXT_TYPE_INVALID,
 };
 
 struct ofono_gprs {
@@ -66,6 +73,8 @@ struct ofono_gprs {
 	struct ofono_netreg *netreg;
 	unsigned int netreg_watch;
 	unsigned int status_watch;
+	GKeyFile *settings;
+	char *imsi;
 	DBusMessage *pending;
 	struct ofono_gprs_context *context_driver;
 	const struct ofono_gprs_driver *driver;
@@ -85,15 +94,16 @@ struct pri_context {
 	ofono_bool_t active;
 	ofono_bool_t pending_active;
 	enum gprs_context_type type;
-	char *name;
+	char name[MAX_CONTEXT_NAME_LENGTH + 1];
 	char *path;
+	char *key;
 	struct ofono_gprs_primary_context context;
 	struct ofono_gprs *gprs;
 };
 
 static void gprs_netreg_update(struct ofono_gprs *gprs);
 
-static inline const char *gprs_context_type_to_string(int type)
+static const char *gprs_context_type_to_string(int type)
 {
 	switch (type) {
 	case GPRS_CONTEXT_TYPE_INTERNET:
@@ -107,6 +117,18 @@ static inline const char *gprs_context_type_to_string(int type)
 	return NULL;
 }
 
+static enum gprs_context_type gprs_context_string_to_type(const char *str)
+{
+	if (g_str_equal(str, "internet"))
+		return GPRS_CONTEXT_TYPE_INTERNET;
+	else if (g_str_equal(str, "wap"))
+		return GPRS_CONTEXT_TYPE_WAP;
+	else if (g_str_equal(str, "mms"))
+		return GPRS_CONTEXT_TYPE_MMS;
+
+	return GPRS_CONTEXT_TYPE_INVALID;
+}
+
 static struct pri_context *gprs_context_by_path(struct ofono_gprs *gprs,
 						const char *ctx_path)
 {
@@ -131,7 +153,7 @@ static DBusMessage *pri_get_properties(DBusConnection *conn,
 	DBusMessageIter dict;
 	dbus_bool_t value;
 	const char *type = gprs_context_type_to_string(ctx->type);
-	const char *name = ctx->name ? ctx->name : "";
+	const char *name = ctx->name;
 	const char *strvalue;
 
 	reply = dbus_message_new_method_return(msg);
@@ -202,6 +224,8 @@ static void pri_set_active_callback(const struct ofono_error *error,
 static DBusMessage *pri_set_apn(struct pri_context *ctx, DBusConnection *conn,
 				DBusMessage *msg, const char *apn)
 {
+	GKeyFile *settings = ctx->gprs->settings;
+
 	if (strlen(apn) > OFONO_GPRS_MAX_APN_LENGTH)
 		return __ofono_error_invalid_format(msg);
 
@@ -213,6 +237,12 @@ static DBusMessage *pri_set_apn(struct pri_context *ctx, DBusConnection *conn,
 
 	strcpy(ctx->context.apn, apn);
 
+	if (settings) {
+		g_key_file_set_string(settings, ctx->key,
+					"AccessPointName", apn);
+		storage_sync(ctx->gprs->imsi, SETTINGS_STORE, settings);
+	}
+
 	g_dbus_send_reply(conn, msg, DBUS_TYPE_INVALID);
 
 	ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -227,6 +257,8 @@ static DBusMessage *pri_set_username(struct pri_context *ctx,
 					DBusConnection *conn, DBusMessage *msg,
 					const char *username)
 {
+	GKeyFile *settings = ctx->gprs->settings;
+
 	if (strlen(username) > OFONO_GPRS_MAX_USERNAME_LENGTH)
 		return __ofono_error_invalid_format(msg);
 
@@ -235,6 +267,12 @@ static DBusMessage *pri_set_username(struct pri_context *ctx,
 
 	strcpy(ctx->context.username, username);
 
+	if (settings) {
+		g_key_file_set_string(settings, ctx->key,
+					"Username", username);
+		storage_sync(ctx->gprs->imsi, SETTINGS_STORE, settings);
+	}
+
 	g_dbus_send_reply(conn, msg, DBUS_TYPE_INVALID);
 
 	ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -249,6 +287,8 @@ static DBusMessage *pri_set_password(struct pri_context *ctx,
 					DBusConnection *conn, DBusMessage *msg,
 					const char *password)
 {
+	GKeyFile *settings = ctx->gprs->settings;
+
 	if (strlen(password) > OFONO_GPRS_MAX_PASSWORD_LENGTH)
 		return __ofono_error_invalid_format(msg);
 
@@ -257,6 +297,12 @@ static DBusMessage *pri_set_password(struct pri_context *ctx,
 
 	strcpy(ctx->context.password, password);
 
+	if (settings) {
+		g_key_file_set_string(settings, ctx->key,
+					"Password", password);
+		storage_sync(ctx->gprs->imsi, SETTINGS_STORE, settings);
+	}
+
 	g_dbus_send_reply(conn, msg, DBUS_TYPE_INVALID);
 
 	ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -270,15 +316,12 @@ static DBusMessage *pri_set_password(struct pri_context *ctx,
 static DBusMessage *pri_set_type(struct pri_context *ctx, DBusConnection *conn,
 					DBusMessage *msg, const char *type)
 {
+	GKeyFile *settings = ctx->gprs->settings;
 	enum gprs_context_type context_type;
 
-	if (g_str_equal(type, "internet"))
-		context_type = GPRS_CONTEXT_TYPE_INTERNET;
-	else if (g_str_equal(type, "wap"))
-		context_type = GPRS_CONTEXT_TYPE_WAP;
-	else if (g_str_equal(type, "mms"))
-		context_type = GPRS_CONTEXT_TYPE_MMS;
-	else
+	context_type = gprs_context_string_to_type(type);
+
+	if (context_type == GPRS_CONTEXT_TYPE_INVALID)
 		return __ofono_error_invalid_args(msg);
 
 	if (ctx->type == context_type)
@@ -286,6 +329,11 @@ static DBusMessage *pri_set_type(struct pri_context *ctx, DBusConnection *conn,
 
 	ctx->type = context_type;
 
+	if (settings) {
+		g_key_file_set_string(settings, ctx->key, "Type", type);
+		storage_sync(ctx->gprs->imsi, SETTINGS_STORE, settings);
+	}
+
 	g_dbus_send_reply(conn, msg, DBUS_TYPE_INVALID);
 
 	ofono_dbus_signal_property_changed(conn, ctx->path,
@@ -298,13 +346,20 @@ static DBusMessage *pri_set_type(struct pri_context *ctx, DBusConnection *conn,
 static DBusMessage *pri_set_name(struct pri_context *ctx, DBusConnection *conn,
 					DBusMessage *msg, const char *name)
 {
+	GKeyFile *settings = ctx->gprs->settings;
+
+	if (strlen(name) > MAX_CONTEXT_NAME_LENGTH)
+		return __ofono_error_invalid_format(msg);
+
 	if (ctx->name && g_str_equal(ctx->name, name))
 		return dbus_message_new_method_return(msg);
 
-	if (ctx->name)
-		g_free(ctx->name);
+	strcpy(ctx->name, name);
 
-	ctx->name = g_strdup(name);
+	if (settings) {
+		g_key_file_set_string(settings, ctx->key, "Name", ctx->name);
+		storage_sync(ctx->gprs->imsi, SETTINGS_STORE, settings);
+	}
 
 	g_dbus_send_reply(conn, msg, DBUS_TYPE_INVALID);
 
@@ -435,7 +490,9 @@ static GDBusSignalTable context_signals[] = {
 	{ }
 };
 
-static struct pri_context *pri_context_create(struct ofono_gprs *gprs)
+static struct pri_context *pri_context_create(struct ofono_gprs *gprs,
+						const char *name,
+						enum gprs_context_type type)
 {
 	struct pri_context *context = g_try_new0(struct pri_context, 1);
 
@@ -443,6 +500,8 @@ static struct pri_context *pri_context_create(struct ofono_gprs *gprs)
 		return NULL;
 
 	context->gprs = gprs;
+	strcpy(context->name, name);
+	context->type = type;
 
 	return context;
 }
@@ -451,9 +510,6 @@ static void pri_context_destroy(gpointer userdata)
 {
 	struct pri_context *ctx = userdata;
 
-	if (ctx->name)
-		g_free(ctx->name);
-
 	if (ctx->path)
 		g_free(ctx->path);
 
@@ -465,9 +521,11 @@ static gboolean context_dbus_register(struct pri_context *ctx)
 	DBusConnection *conn = ofono_dbus_get_connection();
 	char path[256];
 	unsigned int id = ctx->gprs->next_context_id;
+	const char *basepath;
 
-	snprintf(path, sizeof(path), "%s/primarycontext%u",
-			__ofono_atom_get_path(ctx->gprs->atom), id);
+	basepath = __ofono_atom_get_path(ctx->gprs->atom);
+
+	snprintf(path, sizeof(path), "%s/primarycontext%u", basepath, id);
 
 	if (!g_dbus_register_interface(conn, path, DATA_CONTEXT_INTERFACE,
 					context_methods, context_signals,
@@ -479,6 +537,7 @@ static gboolean context_dbus_register(struct pri_context *ctx)
 	}
 
 	ctx->path = g_strdup(path);
+	ctx->key = ctx->path + strlen(basepath) + 1;
 
 	ctx->gprs->next_context_id += 1;
 
@@ -692,6 +751,14 @@ static DBusMessage *gprs_set_property(DBusConnection *conn,
 
 		gprs->roaming_allowed = value;
 
+		if (gprs->settings) {
+			g_key_file_set_integer(gprs->settings, SETTINGS_GROUP,
+						"RoamingAllowed",
+						gprs->roaming_allowed);
+			storage_sync(gprs->imsi, SETTINGS_STORE,
+					gprs->settings);
+		}
+
 		gprs_netreg_update(gprs);
 	} else if (!strcmp(property, "Powered")) {
 		if (!gprs->driver->set_attached)
@@ -707,6 +774,13 @@ static DBusMessage *gprs_set_property(DBusConnection *conn,
 
 		gprs->powered = value;
 
+		if (gprs->settings) {
+			g_key_file_set_integer(gprs->settings, SETTINGS_GROUP,
+						"Powered", gprs->powered);
+			storage_sync(gprs->imsi, SETTINGS_STORE,
+					gprs->settings);
+		}
+
 		gprs_netreg_update(gprs);
 	} else
 		return __ofono_error_invalid_args(msg);
@@ -724,13 +798,29 @@ static DBusMessage *gprs_create_context(DBusConnection *conn,
 {
 	struct ofono_gprs *gprs = data;
 	struct pri_context *context;
+	const char *name;
+	const char *typestr;
 	const char *path;
+	enum gprs_context_type type;
 	char **objpath_list;
 
-	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_INVALID))
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &name,
+					DBUS_TYPE_STRING, &typestr,
+					DBUS_TYPE_INVALID))
 		return __ofono_error_invalid_args(msg);
 
-	context = pri_context_create(gprs);
+	if (strlen(name) == 0 || strlen(name) > MAX_CONTEXT_NAME_LENGTH)
+		return __ofono_error_invalid_format(msg);
+
+	type = gprs_context_string_to_type(typestr);
+
+	if (type == GPRS_CONTEXT_TYPE_INVALID)
+		return __ofono_error_invalid_format(msg);
+
+	if (g_slist_length(gprs->contexts) >= MAX_CONTEXTS)
+		return __ofono_error_not_supported(msg);
+
+	context = pri_context_create(gprs, name, type);
 
 	if (!context) {
 		ofono_error("Unable to allocate context struct");
@@ -744,6 +834,21 @@ static DBusMessage *gprs_create_context(DBusConnection *conn,
 		return __ofono_error_failed(msg);
 	}
 
+	if (gprs->settings) {
+		g_key_file_set_string(gprs->settings, context->key,
+					"Name", context->name);
+		g_key_file_set_string(gprs->settings, context->key,
+					"AccessPointName",
+					context->context.apn);
+		g_key_file_set_string(gprs->settings, context->key,
+					"Username", context->context.username);
+		g_key_file_set_string(gprs->settings, context->key,
+					"Password", context->context.password);
+		g_key_file_set_string(gprs->settings, context->key, "Type",
+				gprs_context_type_to_string(context->type));
+		storage_sync(gprs->imsi, SETTINGS_STORE, gprs->settings);
+	}
+
 	gprs->contexts = g_slist_append(gprs->contexts, context);
 
 	objpath_list = gprs_contexts_path_list(gprs->contexts);
@@ -780,6 +885,11 @@ static void gprs_deactivate_for_remove(const struct ofono_error *error,
 		return;
 	}
 
+	if (gprs->settings) {
+		g_key_file_remove_group(gprs->settings, ctx->key, NULL);
+		storage_sync(gprs->imsi, SETTINGS_STORE, gprs->settings);
+	}
+
 	context_dbus_unregister(ctx);
 	gprs->contexts = g_slist_remove(gprs->contexts, ctx);
 
@@ -828,6 +938,11 @@ static DBusMessage *gprs_remove_context(DBusConnection *conn,
 		return NULL;
 	}
 
+	if (gprs->settings) {
+		g_key_file_remove_group(gprs->settings, ctx->key, NULL);
+		storage_sync(gprs->imsi, SETTINGS_STORE, gprs->settings);
+	}
+
 	ofono_debug("Unregistering context: %s\n", ctx->path);
 	context_dbus_unregister(ctx);
 	gprs->contexts = g_slist_remove(gprs->contexts, ctx);
@@ -865,7 +980,7 @@ static DBusMessage *gprs_deactivate_all(DBusConnection *conn,
 static GDBusMethodTable manager_methods[] = {
 	{ "GetProperties",	"",	"a{sv}",	gprs_get_properties },
 	{ "SetProperty",	"sv",	"",		gprs_set_property },
-	{ "CreateContext",	"",	"o",		gprs_create_context },
+	{ "CreateContext",	"ss",	"o",		gprs_create_context },
 	{ "RemoveContext",	"o",	"",		gprs_remove_context,
 							G_DBUS_METHOD_FLAG_ASYNC },
 	{ "DeactivateAll",	"",	"",		gprs_deactivate_all,
@@ -1149,6 +1264,15 @@ static void gprs_unregister(struct ofono_atom *atom)
 	const char *path = __ofono_atom_get_path(atom);
 	GSList *l;
 
+	if (gprs->settings) {
+		storage_close(gprs->imsi, SETTINGS_STORE,
+				gprs->settings, TRUE);
+
+		g_free(gprs->imsi);
+		gprs->imsi = NULL;
+		gprs->settings = NULL;
+	}
+
 	for (l = gprs->contexts; l; l = l->next) {
 		struct pri_context *context = l->data;
 
@@ -1255,12 +1379,157 @@ static void netreg_watch(struct ofono_atom *atom,
 	gprs_netreg_update(gprs);
 }
 
+static gboolean load_context(struct ofono_gprs *gprs, const char *group)
+{
+	char *name = NULL;
+	char *typestr = NULL;
+	char *username = NULL;
+	char *password = NULL;
+	char *apn = NULL;
+	gboolean ret = FALSE;
+	struct pri_context *context;
+	enum gprs_context_type type;
+
+	if ((name = g_key_file_get_string(gprs->settings, group,
+					"Name", NULL)) == NULL)
+		goto err;
+
+	if ((typestr = g_key_file_get_string(gprs->settings, group,
+					"Type", NULL)) == NULL)
+		goto err;
+
+	type = gprs_context_string_to_type(typestr);
+	if (type == GPRS_CONTEXT_TYPE_INVALID)
+		goto err;
+
+	username = g_key_file_get_string(gprs->settings, group,
+						"Username", NULL);
+	if (!username)
+		goto err;
+
+	if (strlen(username) > OFONO_GPRS_MAX_USERNAME_LENGTH)
+		goto err;
+
+	password = g_key_file_get_string(gprs->settings, group,
+						"Password", NULL);
+
+	if (!password)
+		goto err;
+
+	if (strlen(password) > OFONO_GPRS_MAX_PASSWORD_LENGTH)
+		goto err;
+
+	apn = g_key_file_get_string(gprs->settings, group,
+					"AccessPointName", NULL);
+
+	if (!apn)
+		goto err;
+
+	if (strlen(apn) > OFONO_GPRS_MAX_APN_LENGTH)
+		goto err;
+
+	/* Accept empty (just created) APNs, but don't allow other
+	 * invalid ones */
+	if (apn[0] != '\0' && is_valid_apn(apn) == FALSE)
+		goto err;
+
+	if ((context = pri_context_create(gprs, name, type)) == NULL)
+		goto err;
+
+	strcpy(context->context.username, username);
+	strcpy(context->context.password, password);
+	strcpy(context->context.apn, apn);
+
+	if (context_dbus_register(context) == FALSE)
+		goto err;
+
+	gprs->contexts = g_slist_append(gprs->contexts, context);
+	ret = TRUE;
+
+err:
+	if (name)
+		g_free(name);
+
+	if (typestr)
+		g_free(typestr);
+
+	if (username)
+		g_free(username);
+
+	if (password)
+		g_free(password);
+
+	if (apn)
+		g_free(apn);
+
+	return ret;
+}
+
+static void gprs_load_settings(struct ofono_gprs *gprs, const char *imsi)
+{
+	GError *error = NULL;
+	char **groups;
+	int i;
+
+	gprs->settings = storage_open(imsi, SETTINGS_STORE);
+
+	if (gprs->settings == NULL)
+		return;
+
+	gprs->imsi = g_strdup(imsi);
+
+	gprs->powered = g_key_file_get_boolean(gprs->settings, SETTINGS_GROUP,
+						"Powered", &error);
+
+	/*
+	 * If any error occurs, simply switch to defaults.
+	 * Default to Powered = True
+	 * and RoamingAllowed = True
+	 */
+	if (error) {
+		gprs->powered = TRUE;
+		g_key_file_set_boolean(gprs->settings, SETTINGS_GROUP,
+					"Powered", gprs->powered);
+	}
+
+	gprs->roaming_allowed = g_key_file_get_boolean(gprs->settings,
+							SETTINGS_GROUP,
+							"RoamingAllowed", NULL);
+
+	if (error) {
+		gprs->roaming_allowed = TRUE;
+		g_key_file_set_boolean(gprs->settings, SETTINGS_GROUP,
+					"RoamingAllowed",
+					gprs->roaming_allowed);
+	}
+
+	groups = g_key_file_get_groups(gprs->settings, NULL);
+
+	for (i = 0; groups[i]; i++) {
+
+		if (g_str_equal(groups[i], SETTINGS_GROUP))
+			continue;
+
+		if (!g_str_has_prefix(groups[i], "primarycontext"))
+			goto remove;
+
+		if (load_context(gprs, groups[i]) == TRUE)
+			continue;
+
+remove:
+		g_key_file_remove_group(gprs->settings, groups[i], NULL);
+	}
+
+	g_strfreev(groups);
+}
+
 void ofono_gprs_register(struct ofono_gprs *gprs)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	struct ofono_modem *modem = __ofono_atom_get_modem(gprs->atom);
 	const char *path = __ofono_atom_get_path(gprs->atom);
 	struct ofono_atom *netreg_atom;
+	struct ofono_atom *sim_atom;
 
 	if (!g_dbus_register_interface(conn, path,
 					DATA_CONNECTION_MANAGER_INTERFACE,
@@ -1274,8 +1543,14 @@ void ofono_gprs_register(struct ofono_gprs *gprs)
 
 	ofono_modem_add_interface(modem, DATA_CONNECTION_MANAGER_INTERFACE);
 
-	/* TODO: Read Powered from SIM store */
-	gprs->powered = TRUE;
+	sim_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_SIM);
+
+	if (sim_atom) {
+		struct ofono_sim *sim = __ofono_atom_get_data(sim_atom);
+		const char *imsi = ofono_sim_get_imsi(sim);
+
+		gprs_load_settings(gprs, imsi);
+	}
 
 	gprs->netreg_watch = __ofono_modem_add_atom_watch(modem,
 					OFONO_ATOM_TYPE_NETREG,
